# Copyright 2013 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation

"""ctypes bindings for libgpod.

Similar to the in libgpod included swig bindings except:
  - Functions for pixbuf handling are missing (itdb_artwork_*)
  - Swig wrappers (sw_*) from the swig bindings are mostly missing

These are needed since the original bindings depend on pygtk which breaks
in combination with pygobject.
"""

from ctypes import *


try:
    _glib = CDLL("libglib-2.0.so.0")
except OSError:
    raise ImportError("Couldn't find libglib-2.0")


try:
    _lib = CDLL("libgpod.so.4")
except OSError:
    raise ImportError("Couldn't find libgpod")


gchar_p = c_char_p
gchar = c_char
guchar = c_uint8
guchar_p = POINTER(c_uint8)
guint = c_uint
gpointer = c_void_p
gint32 = c_int32
guint32 = c_uint32
gint = c_int
gboolean = c_bool
gint8 = c_int8
guint8 = c_uint8
gint16 = c_int16
guint16 = c_uint16
gint64 = c_int64
guint64 = c_uint64
gfloat = c_float
gdouble = c_double
gshort = c_short
gushort = c_ushort
glong = c_long
gulong = c_ulong
gsize = c_size_t
gconstpointer = c_void_p

time_t = c_ulong


g_strdup = _glib.g_strdup
g_strdup.argtypes = [gchar_p]
g_strdup.restype = gpointer

g_free = _glib.g_free
g_free.argtypes = [gpointer]
g_free.restype = None


class StructPointerMixin(object):
    """Access struct fields through the struct pointer"""

    def _get_type(self, name):
        """Returns the type of the field"""

        for n, type_ in self._fields_:
            if name == n:
                return type_
        else:
            raise KeyError

    def __getattr__(self, name):
        if not self:
            raise AttributeError("NULL pointer access")

        return getattr(self.contents, name)

    def __setattr__(self, name, value):
        if not self:
            raise AttributeError("NULL pointer access")

        if not hasattr(self.contents, name):
            raise AttributeError("Invalid attribute: %r" % name)

        # copy c_char_p values and free old ones
        type_ = self._get_type(name)
        if type_ is gchar_p:
            old_value = getattr(self.contents, name)
            g_free(old_value)
            value = g_strdup(value)

        setattr(self.contents, name, value)


class GList(Structure):
    pass


class GListPtr(POINTER(GList), StructPointerMixin):
    _type_ = GList


GList._fields_ = [
    ("data", gpointer),
    ("next", GListPtr),
    ("prev", GListPtr),
]


class GTreePtr(gpointer):
    pass


class Enum(guint):
    def __str__(self):
        for a in (c for c in dir(self) if c.upper() == c):
            if getattr(self, a) == self.value:
                return a
        return "Unkown"

    def __repr__(self):
        return repr(str(self))

    def __int__(self):
        return self.value


class Flags(guint):
    def __str__(self):
        values = []
        for a in (c for c in dir(self) if c.upper() == c):
            if getattr(self, a) & self.value:
                values.append(a)
        return " | ".join(values) or "Unkown"

    def __repr__(self):
        return repr(str(self))

    def __int__(self):
        return self.value


class IpodGeneration(Enum):
    (UNKNOWN, FIRST, SECOND, THIRD, FOURTH, PHOTO, MOBILE, MINI_1, MINI_2,
    SHUFFLE_1, SHUFFLE_2, SHUFFLE_3, NANO_1, NANO_2, NANO_3, NANO_4, VIDEO_1,
    VIDEO_2, CLASSIC_1, CLASSIC_2, TOUCH_1, IPHONE_1, SHUFFLE_4, TOUCH_2,
    IPHONE_2, IPHONE_3, CLASSIC_3, NANO_5, TOUCH_3, IPAD_1, IPHONE_4,
    TOUCH_4, NANO_6) = range(33)


class IpodModel(Enum):
    (INVALID, UNKNOWN, COLOR, COLOR_U2, REGULAR, REGULAR_U2, MINI, MINI_BLUE,
    MINI_PINK, MINI_GREEN, MINI_GOLD, SHUFFLE, NANO_WHITE, NANO_BLACK,
    VIDEO_WHITE, VIDEO_BLACK, MOBILE_1, VIDEO_U2, NANO_SILVER, NANO_BLUE,
    NANO_GREEN, NANO_PINK, NANO_RED, NANO_YELLOW, NANO_PURPLE, NANO_ORANGE,
    IPHONE_1, SHUFFLE_SILVER, SHUFFLE_PINK, SHUFFLE_BLUE, SHUFFLE_GREEN,
    SHUFFLE_ORANGE, SHUFFLE_PURPLE, SHUFFLE_RED, CLASSIC_SILVER,
    CLASSIC_BLACK, TOUCH_SILVER, SHUFFLE_BLACK, IPHONE_WHITE, IPHONE_BLACK,
    SHUFFLE_GOLD, SHUFFLE_STAINLESS, IPAD) = range(43)


class IpodInfo(Structure):
    _fields_ = [
        ("model_number", gchar_p),
        ("capacity", c_double),
        ("ipod_model", IpodModel),
        ("ipod_generation", IpodGeneration),
        ("musicdirs", guint),
        ("reserved_int1", gint32),
        ("reserved_int2", gint32),
        ("reserved1", gconstpointer),
        ("reserved2", gconstpointer),
    ]


class IpodInfoPtr(POINTER(IpodInfo), StructPointerMixin):
    _type_ = IpodInfo


ITDB_SPL_STRING_MAXLEN = 255
ITDB_SPL_DATE_IDENTIFIER = 0x2dae2dae2dae2dae


class SPLMatch(Enum):
    AND = 0
    OR = 1


class LimitType(Enum):
    MINUTES = 0x01
    MB = 0x02
    SONGS = 0x03
    HOURS = 0x04
    GB = 0x05


class LimitSort(Enum):
    RANDOM = 0x02
    SONG_NAME = 0x03
    ALBUM = 0x04
    ARTIST = 0x05
    GENRE = 0x07
    MOST_RECENTLY_ADDED = 0x10
    LEAST_RECENTLY_ADDED = 0x80000010
    MOST_OFTEN_PLAYED = 0x14
    LEAST_OFTEN_PLAYED = 0x80000014
    MOST_RECENTLY_PLAYED = 0x15
    LEAST_RECENTLY_PLAYED = 0x80000015
    HIGHEST_RATING = 0x17
    LOWEST_RATING = 0x80000017


class SPLAction(Enum):
    IS_INT = 0x00000001
    IS_GREATER_THAN = 0x00000010
    IS_LESS_THAN = 0x00000040
    IS_IN_THE_RANGE = 0x00000100
    IS_IN_THE_LAST = 0x00000200
    BINARY_AND = 0x00000400

    BINARY_UNKNOWN1 = 0x00000800

    IS_STRING = 0x01000001
    CONTAINS = 0x01000002
    STARTS_WITH = 0x01000004
    ENDS_WITH = 0x01000008

    IS_NOT_INT = 0x02000001
    IS_NOT_GREATER_THAN = 0x02000010
    IS_NOT_LESS_THAN = 0x02000040
    IS_NOT_IN_THE_RANGE = 0x02000100
    IS_NOT_IN_THE_LAST = 0x02000200
    NOT_BINARY_AND = 0x02000400
    BINARY_UNKNOWN2 = 0x02000800

    IS_NOT = 0x03000001
    DOES_NOT_CONTAIN = 0x03000002
    DOES_NOT_START_WITH = 0x03000004
    DOES_NOT_END_WITH = 0x03000008


class SPLFieldType(Enum):
    (STRING, INT, BOOLEAN, DATE, PLAYLIST, UNKNOWN, BINARY_AND) = range(1, 8)


class SPLActionType(Enum):
    (STRING, INT, DATE, RANGE_INT, RANGE_DATE, INTHELAST, PLAYLIST, NONE,
    INVALID, UNKNOWN, BINARY_AND) = range(1, 12)


class SPLActionLast(Enum):
    DAYS_VALUE = 86400
    WEEKS_VALUE = 604800
    MONTHS_VALUE = 2628000


class SPLField(Enum):
    SONG_NAME = 0x02
    ALBUM = 0x03
    ARTIST = 0x04
    BITRATE = 0x05
    SAMPLE_RATE = 0x06
    YEAR = 0x07
    GENRE = 0x08
    KIND = 0x09
    DATE_MODIFIED = 0x0a
    TRACKNUMBER = 0x0b
    SIZE = 0x0c
    TIME = 0x0d
    COMMENT = 0x0e
    DATE_ADDED = 0x10
    COMPOSER = 0x12
    PLAYCOUNT = 0x16
    LAST_PLAYED = 0x17
    DISC_NUMBER = 0x18
    RATING = 0x19
    COMPILATION = 0x1f
    BPM = 0x23
    GROUPING = 0x27
    PLAYLIST = 0x28
    PURCHASE = 0x29
    DESCRIPTION = 0x36
    CATEGORY = 0x37
    PODCAST = 0x39
    VIDEO_KIND = 0x3c
    TVSHOW = 0x3e
    SEASON_NR = 0x3f
    SKIPCOUNT = 0x44
    LAST_SKIPPED = 0x45
    ALBUMARTIST = 0x47
    SORT_SONG_NAME = 0x4e
    SORT_ALBUM = 0x4f
    SORT_ARTIST = 0x50
    SORT_ALBUMARTIST = 0x51
    SORT_COMPOSER = 0x52
    SORT_TVSHOW = 0x53
    ALBUM_RATING = 0x5a


class SPLPref(Structure):
    _fields_ = [
        ("liveupdate", guint8),
        ("checkrules", guint8),
        ("checklimits", guint8),
        ("limittype", guint32),
        ("limitsort", guint32),
        ("limitvalue", guint32),
        ("matchcheckedonly", guint8),
        ("reserved_int1", gint32),
        ("reserved_int2", gint32),
        ("reserved1", gpointer),
        ("reserved2", gpointer),
    ]


class SPLRule(Structure):
    _fields_ = [
        ("field", guint32),
        ("action", guint32),
        ("string", gchar_p),
        ("fromvalue", guint64),
        ("fromdate", guint64),
        ("fromunits", guint64),
        ("tovalue", guint64),
        ("todate", gint64),
        ("tounits", guint64),
        ("unk052", guint32),
        ("unk056", guint32),
        ("unk060", guint32),
        ("unk064", guint32),
        ("unk068", guint32),
        ("reserved_int1", gint32),
        ("reserved_int2", gint32),
        ("reserved1", gpointer),
        ("reserved2", gpointer),
    ]


class SPLRulePtr(POINTER(SPLRule), StructPointerMixin):
    _type_ = SPLRule


class SPLRules(Structure):
    _fields_ = [
        ("unk004", guint32),
        ("match_operator", guint32),
        ("rules", GListPtr),
        ("reserved_int1", gint32),
        ("reserved_int2", gint32),
        ("reserved1", gpointer),
        ("reserved2", gpointer),
    ]


class Chapter(Structure):
    _fields_ = [
        ("startpos", guint32),
        ("chaptertitle", gchar_p),
        ("reserved_int1", gint32),
        ("reserved_int2", gint32),
        ("reserved1", gpointer),
        ("reserved2", gpointer),
    ]


class ChapterPtr(POINTER(Chapter), StructPointerMixin):
    _type_ = Chapter


class Chapterdata(Structure):
    _fields_ = [
        ("chapters", GListPtr),
        ("unk024", guint32),
        ("unk028", guint32),
        ("unk032", guint32),
        ("reserved_int1", gint32),
        ("reserved_int2", gint32),
        ("reserved1", gpointer),
        ("reserved2", gpointer),
    ]


class ChapterdataPtr(POINTER(Chapterdata), StructPointerMixin):
    _type_ = Chapterdata


ITDB_RATING_STEP = 20


class ThumbPtr(c_void_p):
    pass


UserDataDuplicateFunc = CFUNCTYPE(gpointer, gpointer)
UserDataDestroyFunc = CFUNCTYPE(None, gpointer)


class Artwork(Structure):
    _fields_ = [
        ("thumbnail", ThumbPtr),
        ("id", guint32),
        ("dbid", guint64),
        ("unk028", gint32),
        ("rating", guint32),
        ("unk036", gint32),
        ("creation_date", time_t),
        ("digitized_date", time_t),
        ("artwork_size", guint32),
        ("reserved_int1", gint32),
        ("reserved_int2", gint32),
        ("reserved1", gpointer),
        ("reserved2", gpointer),
        ("usertype", guint64),
        ("userdata", gpointer),
        ("userdata_duplicate", UserDataDuplicateFunc),
        ("userdata_destroy", UserDataDestroyFunc),
    ]


class ArtworkPtr(POINTER(Artwork), StructPointerMixin):
    _type_ = Artwork


class PhotoDB(Structure):
    _fields_ = [
        ("photos", GListPtr),
        ("photoalbums", GListPtr),
        ("reserved_int1", gint32),
        ("reserved_int2", gint32),
        ("reserved1", gpointer),
        ("reserved2", gpointer),
        ("usertype", guint64),
        ("userdata", gpointer),
        ("userdata_duplicate", UserDataDuplicateFunc),
        ("userdata_destroy", UserDataDestroyFunc),
    ]


class PhotoDBPtr(POINTER(PhotoDB), StructPointerMixin):
    _type_ = PhotoDB


class DevicePtr(c_void_p):
    pass


class iTunesDB_PrivatePtr(c_void_p):
    pass


class iTunesDB(Structure):
    _fields_ = [
        ("tracks", GListPtr),
        ("playlists", GListPtr),
        ("filename", gchar_p),
        ("device", DevicePtr),
        ("version", guint32),
        ("id", guint64),
        ("tzoffset", gint32),
        ("reserved_int2", gint32),
        ("priv", iTunesDB_PrivatePtr),
        ("reserved2", gpointer),
        ("usertype", guint64),
        ("userdata", gpointer),
        ("userdata_duplicate", UserDataDuplicateFunc),
        ("userdata_destroy", UserDataDestroyFunc),
    ]


class iTunesDBPtr(POINTER(iTunesDB), StructPointerMixin):
    _type_ = iTunesDB


class PhotoAlbum(Structure):
    _fields_ = [
        ("photodb", PhotoDBPtr),
        ("name", gchar_p),
        ("members", GListPtr),
        ("album_type", guint8),
        ("playmusic", guint8),
        ("repeat", guint8),
        ("random", guint8),
        ("show_titles", guint8),
        ("transition_direction", guint8),
        ("slide_duration", gint32),
        ("transition_duration", gint32),
        ("song_id", gint64),
        ("unk024", gint32),
        ("unk028", gint16),
        ("unk044", gint32),
        ("unk048", gint32),
        ("album_id", gint32),
        ("prev_album_id", gint32),
        ("reserved_int1", gint32),
        ("reserved_int2", gint32),
        ("reserved1", gpointer),
        ("reserved2", gpointer),
        ("usertype", guint64),
        ("userdata", gpointer),
        ("userdata_duplicate", UserDataDuplicateFunc),
        ("userdata_destroy", UserDataDestroyFunc),
    ]


class PhotoAlbumPtr(POINTER(PhotoAlbum), StructPointerMixin):
    _type_ = PhotoAlbum


class Playlist_PrivatePtr(c_void_p):
    pass


class Playlist(Structure):
    _fields_ = [
        ("itdb", iTunesDBPtr),
        ("name", gchar_p),
        ("type", guint8),
        ("flag1", guint8),
        ("flag2", guint8),
        ("flag3", guint8),
        ("num", gint),
        ("members", GListPtr),
        ("is_spl", gboolean),
        ("timestamp", time_t),
        ("id", guint64),
        ("sortorder", guint32),
        ("podcastflag", guint32),
        ("splpref", SPLPref),
        ("splrules", SPLRules),
        ("reserved100", gpointer),
        ("reserved101", gpointer),
        ("reserved_int1", gint32),
        ("reserved_int2", gint32),
        ("*priv", Playlist_PrivatePtr),
        ("reserved2", gpointer),
        ("usertype", guint64),
        ("userdata", gpointer),
        ("userdata_duplicate", UserDataDuplicateFunc),
        ("userdata_destroy", UserDataDestroyFunc),
    ]


class PlaylistPtr(POINTER(Playlist), StructPointerMixin):
    _type_ = Playlist


class PlaylistSortOrder(Enum):
    MANUAL = 1
    TITLE = 3
    ALBUM = 4
    ARTIST = 5
    BITRATE = 6
    GENRE = 7
    FILETYPE = 8
    TIME_MODIFIED = 9
    TRACK_NR = 10
    SIZE = 11
    TIME = 12
    YEAR = 13
    SAMPLERATE = 14
    COMMENT = 15
    TIME_ADDED = 16
    EQUALIZER = 17
    COMPOSER = 18
    PLAYCOUNT = 20
    TIME_PLAYED = 21
    CD_NR = 22
    RATING = 23
    RELEASE_DATE = 24
    BPM = 25
    GROUPING = 26
    CATEGORY = 27
    DESCRIPTION = 28


class Mediatype(Flags):
    AUDIO = 1 << 0
    MOVIE = 1 << 1
    PODCAST = 1 << 2
    AUDIOBOOK = 1 << 3
    MUSICVIDEO = 1 << 5
    TVSHOW = 1 << 6
    RINGTONE = 1 << 14
    RENTAL = 1 << 15
    ITUNES_EXTRA = 1 << 16
    MEMO = 1 << 20
    ITUNES_U = 1 << 21
    EPUB_BOOK = 1 << 22
    PDF_BOOK = 1 << 23


class Track_PrivatePtr(c_void_p):
    pass


class Track(Structure):
    _fields_ = [
        ("itdb", iTunesDBPtr),
        ("title", gchar_p),
        ("ipod_path", gchar_p),
        ("album", gchar_p),
        ("artist", gchar_p),
        ("genre", gchar_p),
        ("filetype", gchar_p),
        ("comment", gchar_p),
        ("category", gchar_p),
        ("composer", gchar_p),
        ("grouping", gchar_p),
        ("description", gchar_p),
        ("podcasturl", gchar_p),
        ("podcastrss", gchar_p),
        ("chapterdata", ChapterdataPtr),
        ("subtitle", gchar_p),
        ("tvshow", gchar_p),
        ("tvepisode", gchar_p),
        ("tvnetwork", gchar_p),
        ("albumartist", gchar_p),
        ("keywords", gchar_p),
        ("sort_artist", gchar_p),
        ("sort_title", gchar_p),
        ("sort_album", gchar_p),
        ("sort_albumartist", gchar_p),
        ("sort_composer", gchar_p),
        ("sort_tvshow", gchar_p),
        ("id", guint32),
        ("size", guint32),
        ("tracklen", gint32),
        ("cd_nr", gint32),
        ("cds", gint32),
        ("track_nr", gint32),
        ("tracks", gint32),
        ("bitrate", gint32),
        ("samplerate", guint16),
        ("samplerate_low", guint16),
        ("year", gint32),
        ("volume", gint32),
        ("soundcheck", guint32),
        ("time_added", time_t),
        ("time_modified", time_t),
        ("time_played", time_t),
        ("bookmark_time", guint32),
        ("rating", guint32),
        ("playcount", guint32),
        ("playcount2", guint32),
        ("recent_playcount", guint32),
        ("transferred", gboolean),
        ("BPM", gint16),
        ("app_rating", guint8),
        ("type1", guint8),
        ("type2", guint8),
        ("compilation", guint8),
        ("starttime", guint32),
        ("stoptime", guint32),
        ("checked", guint8),
        ("dbid", guint64),
        ("drm_userid", guint32),
        ("visible", guint32),
        ("filetype_marker", guint32),
        ("artwork_count", guint16),
        ("artwork_size", guint32),
        ("samplerate2", c_float),
        ("unk126", guint16),
        ("unk132", guint32),
        ("time_released", time_t),
        ("unk144", guint16),
        ("explicit_flag", guint16),
        ("unk148", guint32),
        ("unk152", guint32),
        ("skipcount", guint32),
        ("recent_skipcount", guint32),
        ("last_skipped", guint32),
        ("has_artwork", guint8),
        ("skip_when_shuffling", guint8),
        ("remember_playback_position", guint8),
        ("flag4", guint8),
        ("dbid2", guint64),
        ("lyrics_flag", guint8),
        ("movie_flag", guint8),
        ("mark_unplayed", guint8),
        ("unk179", guint8),
        ("unk180", guint32),
        ("pregap", guint32),
        ("samplecount", guint64),
        ("unk196", guint32),
        ("postgap", guint32),
        ("unk204", guint32),
        ("mediatype", guint32),
        ("season_nr", guint32),
        ("episode_nr", guint32),
        ("unk220", guint32),
        ("unk224", guint32),
        ("unk228", guint32),
        ("unk232", guint32),
        ("unk236", guint32),
        ("unk240", guint32),
        ("unk244", guint32),
        ("gapless_data", guint32),
        ("unk252", guint32),
        ("gapless_track_flag", guint16),
        ("gapless_album_flag", guint16),
        ("obsolete", guint16),
        ("artwork", ArtworkPtr),
        ("mhii_link", guint32),
        ("reserved_int1", gint32),
        ("reserved_int2", gint32),
        ("reserved_int3", gint32),
        ("reserved_int4", gint32),
        ("reserved_int5", gint32),
        ("reserved_int6", gint32),
        ("priv", Track_PrivatePtr),
        ("reserved2", gpointer),
        ("reserved3", gpointer),
        ("reserved4", gpointer),
        ("reserved5", gpointer),
        ("reserved6", gpointer),
        ("usertype", guint64),
        ("userdata", gpointer),
        ("userdata_duplicate", UserDataDuplicateFunc),
        ("userdata_destroy", UserDataDestroyFunc),
    ]


class TrackPtr(POINTER(Track), StructPointerMixin):
    _type_ = Track


class FileError(Enum):
    SEEK, CORRUPT, NOTFOUND, RENAME, ITDB_CORRUPT = range(5)


class Error(Enum):
    SEEK, CORRUPT, NOTFOUND, RENAME, ITDB_CORRUPT, SQLITE = range(6)


class GErrorPtrPtr(c_void_p):
    pass


_functions = [
    ("itdb_parse", iTunesDBPtr, [gchar_p, GErrorPtrPtr]),
    ("itdb_parse_file", iTunesDBPtr, [gchar_p, GErrorPtrPtr]),
    ("itdb_write", gboolean, [iTunesDBPtr, GErrorPtrPtr]),
    ("itdb_write_file", gboolean, [iTunesDBPtr, gchar_p, GErrorPtrPtr]),
    ("itdb_shuffle_write", gboolean, [iTunesDBPtr, GErrorPtrPtr]),
    ("itdb_shuffle_write_file", gboolean,
     [iTunesDBPtr, gchar_p, GErrorPtrPtr]),
    ("itdb_start_sync", gboolean, [iTunesDBPtr]),
    ("itdb_stop_sync", gboolean, [iTunesDBPtr]),
    ("itdb_new", iTunesDBPtr, []),
    ("itdb_free", None, [iTunesDBPtr]),
    ("itdb_duplicate", iTunesDBPtr, [iTunesDBPtr]),
    ("itdb_tracks_number", guint32, [iTunesDBPtr]),
    ("itdb_tracks_number_nontransferred", guint32, [iTunesDBPtr]),
    ("itdb_playlists_number", guint32, [iTunesDBPtr]),

    ("itdb_musicdirs_number", gint, [iTunesDBPtr]),
    ("itdb_resolve_path", gchar_p, [gchar_p, POINTER(gchar_p)]),
    ("itdb_rename_files", gboolean, [gchar_p, GErrorPtrPtr]),
    ("itdb_cp_get_dest_filename", gchar_p,
     [TrackPtr, gchar_p, gchar_p, GErrorPtrPtr]),
    ("itdb_cp", gboolean, [gchar_p, gchar_p, GErrorPtrPtr]),
    ("itdb_cp_finalize", TrackPtr, [TrackPtr, gchar_p, gchar_p, GErrorPtrPtr]),
    ("itdb_cp_track_to_ipod", gboolean, [TrackPtr, gchar_p, GErrorPtrPtr]),
    ("itdb_filename_fs2ipod", None, [gchar_p]),
    ("itdb_filename_ipod2fs", None, [gchar_p]),
    ("itdb_filename_on_ipod", gchar_p, [TrackPtr]),
    ("itdb_set_mountpoint", None, [iTunesDBPtr, gchar_p]),
    ("itdb_get_mountpoint", gchar_p, [iTunesDBPtr]),
    ("itdb_get_control_dir", gchar_p, [gchar_p]),
    ("itdb_get_itunes_dir", gchar_p, [gchar_p]),
    ("itdb_get_music_dir", gchar_p, [gchar_p]),
    ("itdb_get_artwork_dir", gchar_p, [gchar_p]),
    ("itdb_get_photos_dir", gchar_p, [gchar_p]),
    ("itdb_get_photos_thumb_dir", gchar_p, [gchar_p]),
    ("itdb_get_device_dir", gchar_p, [gchar_p]),
    ("itdb_get_itunesdb_path", gchar_p, [gchar_p]),
    ("itdb_get_itunescdb_path", gchar_p, [gchar_p]),
    ("itdb_get_itunessd_path", gchar_p, [gchar_p]),
    ("itdb_get_artworkdb_path", gchar_p, [gchar_p]),
    ("itdb_get_photodb_path", gchar_p, [gchar_p]),
    ("itdb_get_path", gchar_p, [gchar_p, gchar_p]),

    ("itdb_device_new", DevicePtr, []),
    ("itdb_device_free", None, [DevicePtr]),
    ("itdb_device_set_mountpoint", None, [DevicePtr, gchar_p]),
    ("itdb_device_read_sysinfo", gboolean, [DevicePtr]),
    ("itdb_device_write_sysinfo", gboolean, [DevicePtr, GErrorPtrPtr]),
    ("itdb_device_get_sysinfo", gchar_p, [DevicePtr, gchar_p]),
    ("itdb_device_set_sysinfo", None, [DevicePtr, gchar_p, gchar_p]),
    ("itdb_device_get_ipod_info", IpodInfoPtr, [DevicePtr]),
    ("itdb_info_get_ipod_info_table", IpodInfoPtr, []),
    ("itdb_device_supports_artwork", gboolean, [DevicePtr]),
    ("itdb_device_supports_chapter_image", gboolean, [DevicePtr]),
    ("itdb_device_supports_video", gboolean, [DevicePtr]),
    ("itdb_device_supports_photo", gboolean, [DevicePtr]),
    ("itdb_device_supports_podcast", gboolean, [DevicePtr]),
    ("itdb_info_get_ipod_model_name_string", gchar_p, [IpodModel]),
    ("itdb_info_get_ipod_generation_string", gchar_p, [IpodGeneration]),
    ("itdb_device_get_uuid", gchar_p, [DevicePtr]),

    ("itdb_track_new", TrackPtr, []),
    ("itdb_track_free", None, [TrackPtr]),
    ("itdb_track_add", None, [iTunesDBPtr, TrackPtr, gint32]),
    ("itdb_track_remove", None, [TrackPtr]),
    ("itdb_track_unlink", None, [TrackPtr]),
    ("itdb_track_duplicate", TrackPtr, [TrackPtr]),
    ("itdb_track_by_id", TrackPtr, [iTunesDBPtr, guint32]),
    ("itdb_track_id_tree_create", GTreePtr, [iTunesDBPtr]),
    ("itdb_track_id_tree_destroy", None, [GTreePtr]),
    ("itdb_track_id_tree_by_id", TrackPtr, [GTreePtr, guint32]),

    ("itdb_playlist_new", PlaylistPtr, [gchar_p, gboolean]),
    ("itdb_playlist_free", None, [PlaylistPtr]),
    ("itdb_playlist_add", None, [iTunesDBPtr, PlaylistPtr, gint32]),
    ("itdb_playlist_move", None, [PlaylistPtr, gint32]),
    ("itdb_playlist_remove", None, [PlaylistPtr]),
    ("itdb_playlist_unlink", None, [PlaylistPtr]),
    ("itdb_playlist_duplicate", PlaylistPtr, [PlaylistPtr]),
    ("itdb_playlist_exists", gboolean, [iTunesDBPtr, PlaylistPtr]),
    ("itdb_playlist_add_track", None, [PlaylistPtr, TrackPtr, gint32]),
    ("itdb_playlist_by_id", PlaylistPtr, [iTunesDBPtr, guint64]),
    ("itdb_playlist_by_nr", PlaylistPtr, [iTunesDBPtr, guint32]),
    ("itdb_playlist_by_name", PlaylistPtr, [iTunesDBPtr, gchar_p]),
    ("itdb_playlist_contains_track", gboolean, [PlaylistPtr, TrackPtr]),
    ("itdb_playlist_contain_track_number", guint32, [TrackPtr]),
    ("itdb_playlist_remove_track", None, [PlaylistPtr, TrackPtr]),
    ("itdb_playlist_tracks_number", guint32, [PlaylistPtr]),
    ("itdb_playlist_randomize", None, [PlaylistPtr]),

    ("itdb_playlist_mpl", PlaylistPtr, [iTunesDBPtr]),
    ("itdb_playlist_is_mpl", gboolean, [PlaylistPtr]),
    ("itdb_playlist_set_mpl", None, [PlaylistPtr]),

    ("itdb_playlist_podcasts", PlaylistPtr, [iTunesDBPtr]),
    ("itdb_playlist_is_podcasts", gboolean, [PlaylistPtr]),
    ("itdb_playlist_set_podcasts", None, [PlaylistPtr]),

    ("itdb_playlist_is_audiobooks", gboolean, [PlaylistPtr]),

    ("itdb_splr_get_field_type", SPLFieldType, [SPLRulePtr]),
    ("itdb_splr_get_action_type", SPLActionType, [SPLRulePtr]),
    ("itdb_splr_validate", None, [SPLRulePtr]),
    ("itdb_splr_remove", None, [PlaylistPtr, SPLRulePtr]),
    ("itdb_splr_new", SPLRulePtr, []),
    ("itdb_splr_add", None, [PlaylistPtr, SPLRulePtr, gint]),
    ("itdb_splr_add_new", SPLRulePtr, [PlaylistPtr, gint]),
    ("itdb_spl_copy_rules", None, [PlaylistPtr, PlaylistPtr]),
    ("itdb_splr_eval", gboolean, [SPLRulePtr, TrackPtr]),
    ("itdb_spl_update", None, [PlaylistPtr]),
    ("itdb_spl_update_all", None, [iTunesDBPtr]),
    ("itdb_spl_update_live", None, [iTunesDBPtr]),

    ("itdb_track_set_thumbnails", gboolean, [TrackPtr, gchar_p]),
    ("itdb_track_set_thumbnails_from_data", gboolean,
     [TrackPtr, guchar_p, gsize]),
    ("itdb_track_set_thumbnails_from_pixbuf", gboolean, [TrackPtr, gpointer]),
    ("itdb_track_has_thumbnails", gboolean, [TrackPtr]),
    ("itdb_track_remove_thumbnails", None, [TrackPtr]),
    ("itdb_track_get_thumbnail", gpointer, [TrackPtr, gint, gint]),

    ("itdb_photodb_parse", PhotoDBPtr, [gchar_p, GErrorPtrPtr]),
    ("itdb_photodb_add_photo", ArtworkPtr,
     [PhotoDBPtr, gchar_p, gint, gint, GErrorPtrPtr]),
    ("itdb_photodb_add_photo_from_data", ArtworkPtr,
     [PhotoDBPtr, guchar_p, gsize, gint, gint, GErrorPtrPtr]),
    ("itdb_photodb_add_photo_from_pixbuf", ArtworkPtr,
     [PhotoDBPtr, gpointer, gint, gint, GErrorPtrPtr]),
    ("itdb_photodb_photoalbum_add_photo", None,
     [PhotoDBPtr, PhotoAlbumPtr, ArtworkPtr, gint]),
    ("itdb_photodb_photoalbum_create", PhotoAlbumPtr,
     [PhotoDBPtr, gchar_p, gint]),
    ("itdb_photodb_create", PhotoDBPtr, [gchar_p]),
    ("itdb_photodb_photoalbum_new", PhotoAlbumPtr, [gchar_p]),
    ("itdb_photodb_photoalbum_free", None, [PhotoAlbumPtr]),
    ("itdb_photodb_photoalbum_add", None, [PhotoDBPtr, PhotoAlbumPtr, gint]),
    ("itdb_photodb_free", None, [PhotoDBPtr]),
    ("itdb_photodb_write", gboolean, [PhotoDBPtr, GErrorPtrPtr]),
    ("itdb_photodb_remove_photo", None,
     [PhotoDBPtr, PhotoAlbumPtr, ArtworkPtr]),
    ("itdb_photodb_photoalbum_remove", None,
     [PhotoDBPtr, PhotoAlbumPtr, gboolean]),
    ("itdb_photodb_photoalbum_unlink", None, [PhotoAlbumPtr]),
    ("itdb_photodb_photoalbum_by_name", PhotoAlbumPtr, [PhotoDBPtr, gchar_p]),

    ("itdb_chapterdata_new", ChapterdataPtr, []),
    ("itdb_chapterdata_free", None, [ChapterdataPtr]),
    ("itdb_chapterdata_duplicate", ChapterdataPtr, [ChapterdataPtr]),
    ("itdb_chapterdata_remove_chapter", None, [ChapterdataPtr, ChapterPtr]),
    ("itdb_chapterdata_unlink_chapter", None, [ChapterdataPtr, ChapterPtr]),
    ("itdb_chapterdata_remove_chapters", None, [ChapterdataPtr]),
    ("itdb_chapter_new", ChapterPtr, []),
    ("itdb_chapter_free", None, [ChapterPtr]),
    ("itdb_chapter_duplicate", ChapterPtr, [ChapterPtr]),
    ("itdb_chapterdata_add_chapter", gboolean,
     [ChapterdataPtr, guint32, gchar_p]),

    ("itdb_init_ipod", gboolean, [gchar_p, gchar_p, gchar_p, GErrorPtrPtr]),
]


def wrap_pointer_return(func):
    """If the returned value is a NULL pointer return None instead"""

    def wrapper(*args, **kwargs):
        ret = func(*args, **kwargs)
        if not ret:
            return None
        return ret
    return wrapper


for name, ret, args in _functions:
    try:
        handle = getattr(_lib, name)
    except AttributeError:
        print_d("symbol not found: %r" % name)
        continue
    handle.argtypes = args
    handle.restype = ret

    # validate pointer return values
    if hasattr(ret, "contents"):
        handle = wrap_pointer_return(handle)

    globals()[name] = handle


def sw_get_tracks(itdb_ptr):
    """Get tracks in itdb."""

    if not itdb_ptr:
        raise ValueError

    tracks = []
    node = itdb_ptr.contents.tracks
    while node:
        entry = node.contents
        track_ptr = cast(entry.data, TrackPtr)
        tracks.append(track_ptr)
        node = node.next
    return tracks


__all__ = []
for key in globals().keys():
    lower = key.lower()
    if lower.startswith("itdb_") or lower.startswith("sw_"):
        __all__.append(key)



from tests import TestCase, add

from quodlibet.qltk.msg import *

class _TMessage(TestCase):
    def setUp(self): self.win = self.Kind(None, "title", "description")
    def test_ctr(self): pass
    def tearDown(self): self.win.destroy()

class TWarningMessage(_TMessage): Kind = WarningMessage
add(TWarningMessage)

class TErrorMessage(_TMessage): Kind = ErrorMessage
add(TErrorMessage)

class TConfirmAction(_TMessage): Kind = ConfirmAction
add(TConfirmAction)

class TCancelRevertSave(TestCase):
    def setUp(self): self.win = CancelRevertSave(None)
    def test_ctr(self): pass
    def tearDown(self): self.win.destroy()
add(TCancelRevertSave)



from tests import TestCase, add

from quodlibet.qltk.browser import LibraryBrowser
import quodlibet.config

class TLibraryBrowser(TestCase):
    def setUp(self):
        quodlibet.config.init()

    def test_ctr(self):
        from quodlibet.library import SongLibrary
        from quodlibet.browsers.search import EmptyBar
        win = LibraryBrowser(EmptyBar, SongLibrary())
        win.browser.emit("songs-selected", [], False)
        win.songlist.get_selection().emit("changed")
        win.destroy()

    def tearDown(self):
        quodlibet.config.quit()

add(TLibraryBrowser)



# -*- coding: utf-8 -*-
# Copyright 2007 Javier Kohen, 2010 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation

from tests import TestCase, add

from quodlibet.util import title

class Ttitle(TestCase):
    def test_basics(s):
        s.assertEquals(u"Mama's Boy", title(u"mama's boy"))
        s.assertEquals(u"The A-Sides", title(u"the a-sides"))
        s.assertEquals(u"Hello Goodbye", title(u"hello goodbye"))
        s.assertEquals(u"HELLO GOODBYE", title(u"HELLO GOODBYE"))
        s.assertEquals(u"", title(u""))

    def test_extra_spaces(s):
        s.assertEquals(u"  Space", title(u"  space"))
        s.assertEquals(u" Dodgy  Spaces ", title(u" dodgy  spaces "))

    def test_quirks(s):
        # This character is not an apostrophe, it's a single quote!
        s.assertEquals(u"Mama’S Boy", title(u"mama’s boy"))
        # This is actually an accent character, not an apostrophe either.
        s.assertEquals(u"Mama`S Boy", title(u"mama`s boy"))

    def test_quotes(s):
        s.assertEquals(u"Hello Goodbye (A Song)",
                 title(u"hello goodbye (a song)"))
        s.assertEquals(u"Hello Goodbye 'A Song'",
                 title(u"hello goodbye 'a song'"))
        s.assertEquals(u'Hello Goodbye "A Song"',
                 title(u'hello goodbye "a song"'))
        s.assertEquals(u"Hello Goodbye „A Song”",
                 title(u"hello goodbye „a song”"))
        s.assertEquals(u"Hello Goodbye ‘A Song’",
                 title(u"hello goodbye ‘a song’"))
        s.assertEquals(u"Hello Goodbye “A Song”",
                 title(u"hello goodbye “a song”"))
        s.assertEquals(u"Hello Goodbye »A Song«",
                 title(u"hello goodbye »a song«"))
        s.assertEquals(u"Hello Goodbye «A Song»",
                 title(u"hello goodbye «a song»"))
        s.assertEquals(u"\"24\" Theme",
                 title(u"\"24\" theme"))
        s.assertEquals(u"\"Mad-Dog\" Mike",
                 title(u"\"mad-dog\" mike"))


    def test_unicode(s):
        s.assertEquals(u"Fooäbar",
                 title(u"fooäbar"))
        s.assertEquals(u"Los Años Felices",
                 title(u"los años felices"))
        s.assertEquals(u"Ñandú",
                 title(u"ñandú"))
        s.assertEquals(u"Österreich",
                 title(u"österreich"))
        # Not a real word - there is none with this character at the beginning
        # but still Python doesn't capitalize the es-zed properly.
        # s.assertEquals(u"SSbahn", title(u"ßbahn"))


    # Old tests, somewhat redundant with the above, but you can never have
    # too many tests...

    def test_empty(self):
        self.failUnlessEqual(title(""), "")

    def test_oneword(self):
        self.failUnlessEqual(title("foobar"), "Foobar")

    def test_twowords(self):
        self.failUnlessEqual(title("foo bar"), "Foo Bar")

    def test_preserve(self):
        self.failUnlessEqual(title("fooBar"), "FooBar")

    def test_nonalphabet(self):
        self.failUnlessEqual(title("foo 1bar"), "Foo 1bar")

    def test_two_words_and_one_not(self):
        self.failUnlessEqual(title("foo 1  bar"), "Foo 1  Bar")

    def test_apostrophe(self):
        self.failUnlessEqual(title("it's"), "It's")

add(Ttitle)



# -*- coding: utf-8 -*-
# Copyright 2004-2009 Joe Wreschnig, Michael Urman, Steven Robertson
#           2011,2012 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation

import os
import random
import re
import sys
import traceback
import urlparse
import unicodedata
import urllib
import tempfile
import subprocess
import webbrowser

# title function was moved to a separate module
from quodlibet.util.titlecase import title

from quodlibet.const import FSCODING as fscoding, SUPPORT_EMAIL, COPYRIGHT
from quodlibet.util.dprint import print_d, print_

if os.name == "nt":
    from win32com.shell import shellcon, shell


def strip_win32_incompat(string, BAD='\:*?;"<>|'):
    """Strip Win32-incompatible characters.

    This only works correctly on Unicode strings.
    """
    new = u"".join(map(lambda s: (s in BAD and "_") or s, string))
    parts = new.split(os.sep)

    def fix_end(string):
        return re.sub(r'[\. ]$', "_", string)
    return unicode(os.sep).join(map(fix_end, parts))


def strip_win32_incompat_from_path(string):
    """Strip Win32-incompatible chars from a path, ignoring os.sep
    and the drive part"""
    drive, tail = os.path.splitdrive(string)
    tail = os.sep.join(map(strip_win32_incompat, tail.split(os.sep)))
    return drive + tail


def listdir(path, hidden=False):
    """List files in a directory, sorted, fully-qualified.

    If hidden is false, Unix-style hidden files are not returned.
    """
    path = fsnative(path)
    if hidden:
        filt = None
    else:
        filt = lambda base: not base.startswith(".")
    if path.endswith(os.sep):
        join = "".join
    else:
        join = os.sep.join
    return [join([path, basename])
            for basename in sorted(os.listdir(path))
            if filt(basename)]


class InstanceTracker(object):
    """A mixin for GObjects to return a list of all alive objects
    of a given type. Note that it must be used with a GObject or
    something with a connect method and destroy signal."""
    __kinds = {}

    def _register_instance(self, klass=None):
        """Register this object to be returned in the active instance list."""
        if klass is None:
            klass = type(self)
        self.__kinds.setdefault(klass, []).append(self)
        self.connect('destroy', self.__kinds[klass].remove)

    @classmethod
    def instances(klass):
        return klass.__kinds.get(klass, [])


class OptionParser(object):
    def __init__(self, name, version, description=None, usage=None):
        self.__name = name
        self.__version = version
        self.__args = {}
        self.__translate_short = {}
        self.__translate_long = {}
        self.__help = {}
        self.__usage = usage
        self.__description = description
        self.add(
            "help", shorts="h", help=_("Display brief usage information"))
        self.add(
            "version", shorts="v", help=_("Display version and copyright"))
        self.add("debug", shorts="d")

    def add(self, canon, help=None, arg="", shorts="", longs=[]):
        self.__args[canon] = arg
        for s in shorts:
            self.__translate_short[s] = canon
        for l in longs:
            self.__translate_long[l] = canon
        if help:
            self.__help[canon] = help

    def __shorts(self):
        shorts = ""
        for short, canon in self.__translate_short.items():
            shorts += short + (self.__args[canon] and "=" or "")
        return shorts

    def __longs(self):
        longs = []
        for long, arg in self.__args.items():
            longs.append(long + (arg and "=" or ""))
        for long, canon in self.__translate_long.items():
            longs.append(long + (self.__args[canon] and "=" or ""))
        return longs

    def __format_help(self, opt, space):
        if opt in self.__help:
            help = self.__help[opt]
            if self.__args[opt]:
                opt = "%s=%s" % (opt, self.__args[opt])
            return "  --%s %s\n" % (opt.ljust(space), help)
        else:
            return ""

    def help(self):
        l = 0
        for k in self.__help.keys():
            l = max(l, len(k) + len(self.__args.get(k, "")) + 4)

        s = _("Usage: %s %s\n") % (
            sys.argv[0], self.__usage if self.__usage else _("[options]"))
        if self.__description:
            s += "%s - %s\n" % (self.__name, self.__description)
        s += "\n"
        keys = sorted(self.__help.keys())
        try:
            keys.remove("help")
        except ValueError:
            pass
        try:
            keys.remove("version")
        except ValueError:
            pass
        for h in keys:
            s += self.__format_help(h, l)
        if keys:
            s += "\n"
        s += self.__format_help("help", l)
        s += self.__format_help("version", l)
        return s

    def set_help(self, newhelp):
        self.__help = newhelp

    def version(self):
        return ("""\
{title} {version}
<{email}>
{copyright}\
""").format(title=self.__name, version=self.__version, dates="2004-2012",
            email=SUPPORT_EMAIL, copyright=COPYRIGHT)

    def parse(self, args=None):
        if args is None:
            args = sys.argv[1:]
        from getopt import getopt, GetoptError
        try:
            opts, args = getopt(args, self.__shorts(), self.__longs())
        except GetoptError, s:
            s = str(s)
            text = []
            if "not recognized" in s:
                text.append(
                    _("Option %r not recognized.") % s.split()[1])
            elif "requires argument" in s:
                text.append(
                    _("Option %r requires an argument.") % s.split()[1])
            elif "unique prefix" in s:
                text.append(
                    _("%r is not a unique prefix.") % s.split()[1])
            if "help" in self.__args:
                text.append(_("Try %s --help.") % sys.argv[0])

            print_e("\n".join(text))
            raise SystemExit(True)
        else:
            transopts = {}
            for o, a in opts:
                if o.startswith("--"):
                    o = self.__translate_long.get(o[2:], o[2:])
                elif o.startswith("-"):
                    o = self.__translate_short.get(o[1:], o[1:])
                if o == "help":
                    print_(self.help())
                    raise SystemExit
                elif o == "version":
                    print_(self.version())
                    raise SystemExit
                elif o == "debug":
                    from quodlibet import const
                    const.DEBUG = True
                if self.__args[o]:
                    transopts[o] = a
                else:
                    transopts[o] = True

            return transopts, args


def mtime(filename):
    """Return the mtime of a file, or 0 if an error occurs."""
    try:
        return os.path.getmtime(filename)
    except OSError:
        return 0


def size(filename):
    """Return the size of a file, or 0 if an error occurs."""
    try:
        return os.path.getsize(filename)
    except OSError:
        return 0


def mkdir(dir, *args):
    """Make a directory, including all its parent directories. This does not
    raise an exception if the directory already exists (and is a
    directory)."""
    if not os.path.isdir(dir):
        os.makedirs(dir, *args)


def escape(str):
    """Escape a string in a manner suitable for XML/Pango."""
    return str.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")


def unescape(str):
    """Unescape a string in a manner suitable for XML/Pango."""
    return str.replace("&lt;", "<").replace("&gt;", ">").replace("&amp;", "&")


def escape_filename(s):
    """Escape a string in a manner suitable for a filename."""
    if isinstance(s, unicode):
        s = s.encode("utf-8")
    return urllib.quote(s, safe="").decode("utf-8")


def unescape_filename(s):
    """Unescape a string in a manner suitable for a filename."""
    if isinstance(s, unicode):
        s = s.encode("utf-8")
    return urllib.unquote(s).decode("utf-8")


def parse_time(timestr, err=(ValueError, re.error)):
    """Parse a time string in hh:mm:ss, mm:ss, or ss format."""
    if timestr[0:1] == "-":
        m = -1
        timestr = timestr[1:]
    else:
        m = 1
    try:
        return m * reduce(lambda s, a: s * 60 + int(a),
                          re.split(r":|\.", timestr), 0)
    except err:
        return 0

RATING_PRECISION = 0.25
RATING_SYMBOL = u'\u2605'
BLANK_RATING_SYMBOL = u'\u2606'


def format_rating(value, blank=True):
    """Turn a number into a sequence of rating symbols."""
    steps = int(1 / RATING_PRECISION)
    value = max(min(value, 1.0), 0)
    ons = int(round(steps * value))
    offs = (steps - ons) if blank else 0
    return RATING_SYMBOL * ons + BLANK_RATING_SYMBOL * offs


def format_size(size):
    """Turn an integer size value into something human-readable."""
    # TODO: Better i18n of this (eg use O/KO/MO/GO in French)
    if size >= 1024 ** 3:
        return "%.1f GB" % (float(size) / (1024 ** 3))
    elif size >= 1024 ** 2 * 100:
        return "%.0f MB" % (float(size) / (1024 ** 2))
    elif size >= 1024 ** 2 * 10:
        return "%.1f MB" % (float(size) / (1024 ** 2))
    elif size >= 1024 ** 2:
        return "%.2f MB" % (float(size) / (1024 ** 2))
    elif size >= 1024 * 10:
        return "%d KB" % int(size / 1024)
    elif size >= 1024:
        return "%.2f KB" % (float(size) / 1024)
    else:
        return "%d B" % size


def format_time(time):
    """Turn a time value in seconds into hh:mm:ss or mm:ss."""
    if time < 0:
        time = abs(time)
        prefix = "-"
    else:
        prefix = ""
    if time >= 3600:  # 1 hour
        # time, in hours:minutes:seconds
        return "%s%d:%02d:%02d" % (prefix, time // 3600,
                                   (time % 3600) // 60, time % 60)
    else:
        # time, in minutes:seconds
        return "%s%d:%02d" % (prefix, time // 60, time % 60)


def format_time_long(time):
    """Turn a time value in seconds into x hours, x minutes, etc."""
    if time < 1:
        return _("No time information")
    cutoffs = [
        (60, "%d seconds", "%d second"),
        (60, "%d minutes", "%d minute"),
        (24, "%d hours", "%d hour"),
        (365, "%d days", "%d day"),
        (None, "%d years", "%d year"),
    ]
    time_str = []
    for divisor, plural, single in cutoffs:
        if time < 1:
            break
        if divisor is None:
            time, unit = 0, time
        else:
            time, unit = divmod(time, divisor)
        if unit:
            time_str.append(ngettext(single, plural, unit) % unit)
    time_str.reverse()
    if len(time_str) > 2:
        time_str.pop()
    return ", ".join(time_str)


def fsdecode(s, note=True):
    """Decoding a string according to the filesystem encoding.
    note specifies whether a note should be appended if decoding failed."""
    if isinstance(s, unicode):
        return s
    elif note:
        return decode(s, fscoding)
    else:
        return s.decode(fscoding, 'replace')


def fsencode(s, note=False):
    """Encode a string according to the filesystem encoding.
    note specifies whether a note should be appended if encoding failed."""
    if isinstance(s, str):
        return s
    elif note:
        return encode(s, fscoding)
    else:
        return s.encode(fscoding, 'replace')

"""
Path related functions like open, os.listdir have different behavior on win32

- Passing a string calls the old non unicode win API.
  In case of listdir this leads to "?" for >1byte chars and to
  1 byte chars encoded using the fs encoding. -> DO NOT USE!

- Passing a unicode object internally calls the windows unicode functions.
  This will mostly lead to proper unicode paths (except expanduser).

  And that's why QL is using unicode paths on win and encoded paths
  everywhere else.
"""

if sys.platform == "win32":
    fsnative = fsdecode  # Decode a filename on windows
else:
    fsnative = fsencode  # Encode it on other platforms


def split_scan_dirs(s):
    """Split the value of the "scan" setting, accounting for drive letters on
    win32."""
    if sys.platform == "win32":
        return filter(None, re.findall(r"[a-zA-Z]:[\\/][^:]*", s))
    else:
        return filter(None, s.split(":"))


def decode(s, charset="utf-8"):
    """Decode a string; if an error occurs, replace characters and append
    a note to the string."""
    try:
        return s.decode(charset)
    except UnicodeError:
        return s.decode(charset, "replace") + " " + _("[Invalid Encoding]")


def encode(s, charset="utf-8"):
    """Encode a string; if an error occurs, replace characters and append
    a note to the string."""
    try:
        return s.encode(charset)
    except UnicodeError:
        return (s + " " + _("[Invalid Encoding]")).encode(charset, "replace")


def iscommand(s):
    """True if an executable file 's' exists in the user's path, or is a
    fully-qualified existing executable file."""

    if s == "" or os.path.sep in s:
        return os.path.isfile(s) and os.access(s, os.X_OK)
    else:
        s = s.split()[0]
        for p in os.defpath.split(os.path.pathsep):
            p2 = os.path.join(p, s)
            if os.path.isfile(p2) and os.access(p2, os.X_OK):
                return True
        else:
            return False


def capitalize(str):
    """Capitalize a string, not affecting any character after the first."""
    return str[:1].upper() + str[1:]


def split_value(s, splitters=["/", "&", ","]):
    """Splits a string. The first match in 'splitters' is used as the
    separator; subsequent matches are intentionally ignored."""
    if not splitters:
        return [s.strip()]
    values = s.split("\n")
    for spl in splitters:
        spl = re.compile(r"\b\s*%s\s*\b" % re.escape(spl), re.UNICODE)
        if not filter(spl.search, values):
            continue
        new_values = []
        for v in values:
            new_values.extend([st.strip() for st in spl.split(v)])
        return new_values
    return values


def split_title(s, splitters=["/", "&", ","]):
    title, subtitle = find_subtitle(s)
    return ((title.strip(), split_value(subtitle, splitters))
            if subtitle else (s, []))


__FEATURING = ["feat.", "featuring", "feat", "ft", "ft.", "with", "w/"]
__ORIGINALLY = ["originally by ", " cover"]
# Cache case-insensitive regex searches of the above
__FEAT_REGEX = [re.compile(re.escape(s + " "), re.I) for s in __FEATURING]
__ORIG_REGEX = [re.compile(re.escape(s), re.I) for s in __ORIGINALLY]


def split_people(s, splitters=["/", "&", ","]):
    title, subtitle = find_subtitle(s)
    if not subtitle:
        parts = s.split(" ")
        if len(parts) > 2:
            for feat in __FEATURING:
                try:
                    i = [p.lower() for p in parts].index(feat)
                    orig = " ".join(parts[:i])
                    others = " ".join(parts[i + 1:])
                    return (orig, split_value(others, splitters))
                except (ValueError, IndexError):
                    pass
        return (s, [])
    else:
        old = subtitle
        # TODO: allow multiple substitutions across types, maybe
        for regex in (__FEAT_REGEX + __ORIG_REGEX):
            subtitle = re.sub(regex, "", subtitle, 1)
            if old != subtitle:
                # Only change once
                break
        values = split_value(subtitle, splitters)
        return (title.strip(), values)


def split_album(s):
    name, disc = find_subtitle(s)
    if not disc:
        parts = s.split(" ")
        if len(parts) > 2:
            lower = parts[-2].lower()
            if "disc" in lower or "disk" in lower:
                return (" ".join(parts[:-2]), parts[-1])
        return (s, None)
    else:
        parts = disc.split()
        if (len(parts) == 2 and
                parts[0].lower() in ["disc", "disk", "cd", "vol", "vol."]):
            try:
                return (name, parts[1])
            except:
                return (s, None)
        else:
            return (s, None)


def split_numeric(s, limit=10,
                  reg=re.compile(r"[0-9][0-9]*\.?[0-9]*").search,
                  join=u" ".join):
    """Separate numeric values from the string and convert to float, so
    it can be used for human sorting. Also removes all extra whitespace."""
    result = reg(s)
    if not result or not limit:
        return (join(s.split()),)
    else:
        start, end = result.span()
        return (
            join(s[:start].split()),
            float(result.group()),
            split_numeric(s[end:], limit - 1))


def human_sort_key(s, normalize=unicodedata.normalize):
    if not isinstance(s, unicode):
        s = s.decode("utf-8")
    s = normalize("NFD", s.lower())
    return s and split_numeric(s)


def find_subtitle(title):
    if isinstance(title, str):
        title = title.decode('utf-8', 'replace')
    for pair in [u"[]", u"()", u"~~", u"--", u"\u301c\u301c", u'\uff08\uff09']:
        if pair[0] in title[:-1] and title.endswith(pair[1]):
            r = len(pair[1])
            l = title[0:-r].rindex(pair[0])
            if l != 0:
                subtitle = title[l + len(pair[0]):-r]
                title = title[:l]
                return title.rstrip(), subtitle
    else:
        return title, None


def expanduser(filename):
    """needed because expanduser does not return wide character paths
    on windows even if a unicode path gets passed."""
    if os.name == "nt":
        profile = shell.SHGetFolderPath(0, shellcon.CSIDL_PROFILE, 0, 0)
        if filename == "~":
            return profile
        if filename.startswith(u"~" + os.path.sep):
            return os.path.join(profile, filename[2:])
    return os.path.expanduser(filename)


def unexpand(filename, HOME=expanduser("~")):
    """Replace the user's home directory with ~/, if it appears at the
    start of the path name."""
    sub = (os.name == "nt" and "%USERPROFILE%") or "~"
    if filename == HOME:
        return sub
    elif filename.startswith(HOME + os.path.sep):
        filename = filename.replace(HOME, sub, 1)
    return filename


def website(site):
    """Open the given URL in the user's default browser"""

    if os.name == "nt":
        return webbrowser.open(site)

    # all commands here return immediately
    for prog in ["xdg-open", "gnome-open"]:
        if not iscommand(prog):
            continue

        status = subprocess.check_call([prog, site])
        if status == 0:
            return True

    # sensible-browser is a debian thing
    blocking_progs = ["sensible-browser"]
    blocking_progs.extend(os.environ.get("BROWSER", "").split(":"))

    for prog in blocking_progs:
        if not iscommand(prog):
            continue

        # replace %s with the url
        args = prog.split()
        for i, arg in enumerate(args):
            if arg == "%s":
                args[i] = site
                break
        else:
            args.append(site)

        # calling e.g. firefox blocks, so call async and hope for the best
        try:
            spawn(args)
        except RuntimeError:
            continue
        else:
            return True

    return False


def tag(name, cap=True):
    # Return a 'natural' version of the tag for human-readable bits.
    # Strips ~ and ~# from the start and runs it through a map (which
    # the user can configure).
    if not name:
        return _("Invalid tag")
    else:
        from quodlibet.util.tags import readable
        parts = map(readable, tagsplit(name))
        if cap:
            # Translators: If tag names, when capitalized, should not
            # be title-cased ("Looks Like This"), but rather only have
            # the first letter capitalized, translate this string as
            # something other than "check|titlecase?".
            if _("check|titlecase?") == "check|titlecase?":
                parts = map(title, parts)
            else:
                parts = map(capitalize, parts)
        return " / ".join(parts)


def tagsplit(tag):
    """Split a (potentially) tied tag into a list of atomic tags. Two ~~s
    make the next tag prefixed with a ~, so ~foo~~bar => [foo, ~bar]."""
    if "~" in tag[1:]:
        if tag.startswith("~") and not tag.startswith("~#"):
            tag = tag[1:]
        tags = []
        front = ""
        for part in tag.split("~"):
            if part:
                tags.append(front + part)
                front = ""
            else:
                front = "~"
        return tags
    else:
        return [tag]


def pattern(pat, cap=True, esc=False):
    """Return a 'natural' version of the pattern string for human-readable
    bits. Assumes all tags in the pattern are present."""
    from quodlibet.parse import Pattern, XMLFromPattern

    class Fakesong(dict):
        cap = False

        def comma(self, key):
            return " - ".join(self.list(key))

        def list(self, key):
            return [tag(k, self.cap) for k in tagsplit(key)]
        list_seperate = list
        __call__ = comma

    fakesong = Fakesong({'filename': tag('filename', cap)})
    fakesong.cap = cap
    try:
        p = (esc and XMLFromPattern(pat)) or Pattern(pat)
    except ValueError:
        return _("Invalid pattern")

    return p.format(fakesong)


def spawn(argv, stdout=False):
    """Asynchronously run a program. argv[0] is the executable name, which
    must be fully qualified or in the path. If stdout is True, return
    a file object corresponding to the child's standard output; otherwise,
    return the child's process ID.

    argv must be strictly str objects to avoid encoding confusion.
    """

    from gi.repository import GLib

    types = map(type, argv)
    if not (min(types) == max(types) == str):
        raise TypeError("executables and arguments must be str objects")
    print_d("Running %r" % " ".join(argv))
    args = GLib.spawn_async(argv=argv, flags=GLib.SpawnFlags.SEARCH_PATH,
                            standard_output=stdout)

    if stdout:
        return os.fdopen(args[2])
    else:
        return args[0]


def fver(tup):
    return ".".join(map(str, tup))


def uri_is_valid(uri):
    return bool(urlparse.urlparse(uri)[0])


def make_case_insensitive(filename):
    return "".join(["[%s%s]" % (c.lower(), c.upper()) for c in filename])


def print_exc(limit=None, file=None):
    """A wrapper preventing crashes on broken pipes in print_exc."""
    if not file:
        file = sys.stderr
    print_(traceback.format_exc(limit=limit), output=file)


class DeferredSignal(object):
    """A wrapper for connecting functions to signals.

    Some signals may fire hundreds of times, but only require processing
    once per group. This class pushes the call to the mainloop at idle
    priority and prevents multiple calls from being inserted in the
    mainloop at a time, greatly improving responsiveness in some places.

    Example usage:

    def func(widget, user_arg):
        pass
    widget.connect('signal', DeferredSignal(func), user_arg)
    """

    __slots__ = ['func', 'dirty']

    def __init__(self, func):
        self.func = func
        self.dirty = False

    def __call__(self, *args):
        if not self.dirty:
            self.dirty = True
            from gi.repository import GLib
            GLib.idle_add(self._wrap, *args)

    def _wrap(self, *args):
        self.func(*args)
        self.dirty = False


def gobject_weak(fun, *args, **kwargs):
    """Connect to a signal and disconnect if destroy gets emitted.
    If parent is given, it connects to its destroy signal
    Example:
        gobject_weak(gobject_1.connect, 'changed', self.__changed)
        gobject_weak(gobject_1.connect, 'changed', self.__changed,
            parent=gobject_2)
    """
    parent = kwargs.pop("parent", None)
    obj = fun.__self__
    sig = fun(*args)
    disconnect = lambda obj, handle: obj.disconnect(handle)
    if parent:
        parent.connect_object('destroy', disconnect, obj, sig)
    else:
        obj.connect('destroy', disconnect, sig)
    return sig


class cached_property(object):
    """A read-only @property that is only evaluated once."""
    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


def xdg_get_system_data_dirs():
    """http://standards.freedesktop.org/basedir-spec/latest/"""
    data_dirs = os.getenv("XDG_DATA_DIRS")
    if data_dirs:
        return map(os.path.abspath, data_dirs.split(":"))
    else:
        return ("/usr/local/share/", "/usr/share/")


def xdg_get_cache_home():
    data_home = os.getenv("XDG_CACHE_HOME")
    if data_home:
        return os.path.abspath(data_home)
    else:
        return os.path.join(os.path.expanduser("~"), ".cache")


def xdg_get_data_home():
    data_home = os.getenv("XDG_DATA_HOME")
    if data_home:
        return os.path.abspath(data_home)
    else:
        return os.path.join(os.path.expanduser("~"), ".local", "share")


def find_mount_point(path):
    while not os.path.ismount(path):
        path = os.path.dirname(path)
    return path


def pathname2url_win32(path):
    # stdlib version raises IOError for more than one ':' which can appear
    # using a virtual box shared folder and it inserts /// at the beginning
    # but it should be /.
    quote = urllib.quote
    if path[1:2] != ":" or path[:1] == "\\":
        if path[:2] == "\\\\":
            path = path[2:]
        return quote("/".join(path.split("\\")))
    drive, remain = path.split(":", 1)
    return "/%s:%s" % (quote(drive), quote("/".join(remain.split("\\"))))

if os.name == "nt":
    pathname2url = pathname2url_win32
else:
    pathname2url = urllib.pathname2url


# See http://stackoverflow.com/questions/1151658/python-hashable-dicts
class HashableDict(dict):
    """Standard dict, made hashable. Useful for making sets of dicts etc"""

    def __key(self):
        return tuple((k, self[k]) for k in sorted(self))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return self.__key() == other.__key()


def sanitize_tags(tags, stream=False):
    """Returns a new sanitized tag dict. stream defines if the
    tags of a main/base song should be changed or of a stream song.
    e.g. title will be removed for the base song but not for the stream one.
    """

    san = {}
    for key, value in tags.iteritems():
        key = key.lower()
        key = {"location": "website"}.get(key, key)

        if isinstance(value, unicode):
            lower = value.lower().strip()

            if key == "channel-mode":
                if "stereo" in lower or "dual" in lower:
                    value = u"stereo"
            elif key == "audio-codec":
                if "mp3" in lower:
                    value = u"MP3"
                elif "aac" in lower or "advanced" in lower:
                    value = u"MPEG-4 AAC"
                elif "vorbis" in lower:
                    value = u"Ogg Vorbis"

            if lower in ("http://www.shoutcast.com", "http://localhost/",
                         "default genre", "none", "http://", "unnamed server",
                         "unspecified", "n/a"):
                continue

        if key == "duration":
            try:
                value = int(long(value) / 1000)
            except ValueError:
                pass
            else:
                if not stream:
                    continue
                key = "~#length"
        elif key == "bitrate":
            try:
                value = int(value) / 1000
            except ValueError:
                pass
            else:
                if not stream:
                    continue
                key = "~#bitrate"
        elif key == "nominal-bitrate":
            try:
                value = int(value) / 1000
            except ValueError:
                pass
            else:
                if stream:
                    continue
                key = "~#bitrate"

        if key in ("emphasis", "mode", "layer", "maximum-bitrate",
                   "minimum-bitrate", "has-crc", "homepage"):
            continue

        if not stream and key in ("title", "album", "artist", "date"):
            continue

        if isinstance(value, (int, long, float)):
            if not key.startswith("~#"):
                key = "~#" + key
            san[key] = value
        else:
            if key.startswith("~#"):
                key = key[2:]

            if not isinstance(value, unicode):
                continue

            value = value.strip()
            if key in san:
                if value not in san[key].split("\n"):
                    san[key] += "\n" + value
            else:
                san[key] = value

    return san


def build_filter_query(key, values):
    """Create a text query that matches a union of all values for a key

    build_filter_query("foo", ["x", "y"])
    => foo = |("x"c, "y"c)
    build_filter_query("~#foo", ["1"])
    => #(foo = 1)
    """

    if not values:
        return u""
    if key.startswith("~#"):
        nheader = key[2:]
        queries = ["#(%s = %s)" % (nheader, i) for i in values]
        if len(queries) > 1:
            return u"|(%s)" % ", ".join(queries)
        else:
            return queries[0]
    else:
        text = ", ".join(
            ["'%s'c" % v.replace("\\", "\\\\").replace("'", "\\'")
             for v in values])
        if len(values) == 1:
            return u"%s = %s" % (key, text)
        else:
            return u"%s = |(%s)" % (key, text)


def limit_songs(songs, max, weight_by_ratings=False):
    """Choose at most `max` songs from `songs`,
    optionally giving weighting to ~#rating"""

    if not max or len(songs) < max:
        return songs
    else:
        if weight_by_ratings:
            def choose(r1, r2):
                if r1 or r2:
                    return cmp(random.random(), r1 / (r1 + r2))
                else:
                    return random.randint(-1, 1)

            def rating(song):
                return song("~#rating")
            songs.sort(cmp=choose, key=rating)
        else:
            random.shuffle(songs)
        return songs[:max]


def get_temp_cover_file(data):
    try:
        fn = tempfile.NamedTemporaryFile()
        fn.write(data)
        fn.flush()
        fn.seek(0, 0)
    except EnvironmentError:
        return
    else:
        return fn
